#include "measured.h"

#include <string>
#include <vector>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "esp_log.h"

#include "qthing.h"
#include "mqtt_common.h"


typedef struct {
    std::string name;
    std::string type;
    std::string unit;
    qthing::measured::sensor_t sensor;
} sensor_config_t;

static std::vector<sensor_config_t> sensors;
static qthing::measured::backend_t sens_backend = [](std::vector<qthing::measured::sensor_data_t>& data) {
    for (auto const& d: data) {
        ESP_LOGI("measured", "name=%s type=%s value=%f unit=%s", d.name->c_str(), d.type->c_str(), d.value, d.unit->c_str());
    }
};

static bool measured_running = false;

static void measured(void *ignored) {
    TickType_t lastWakeTime = xTaskGetTickCount();

    while(true) {
        std::vector<qthing::measured::sensor_data_t> data;

        for (auto const& sensor: sensors) {
            qthing::measured::sensor_data_t tmp = {
                .name = &sensor.name,
                .type = &sensor.type,
                .unit = &sensor.unit,
                .value = sensor.sensor()
            };

            data.push_back(tmp);
        }

        sens_backend(data);

        vTaskDelayUntil(&lastWakeTime, MEASURED_PERIOD_MS / portTICK_PERIOD_MS);
    }
}

static void start_measured() {
    // TODO: thread-safe!
    if (!measured_running) {
        measured_running = true;

        // TODO: move somewhere else!
        qthing::measured::registerBackend(qthing::measured::backend::mqtt);

        xTaskCreate(measured, "measured", 8192, NULL, 1, NULL);
    }
}

void qthing::measured::registerSensor(std::string name, std::string type, std::string unit, qthing::measured::sensor_t sensor) {
    sensor_config_t cfg = {
        .name = name,
        .type = type,
        .unit = unit,
        .sensor = sensor
    };

    sensors.push_back(cfg);
    start_measured();
}

void qthing::measured::registerBackend(qthing::measured::backend_t new_backend) {
    qthing::measured::backend_t old_backend = sens_backend;
    
    sens_backend = [old_backend, new_backend](std::vector<qthing::measured::sensor_data_t>& data) {
        old_backend(data);
        new_backend(data);
    };

    start_measured();
}

void qthing::measured::backend::mqtt(std::vector<qthing::measured::sensor_data_t>& data) {
    if (is_mqtt_connected()) {
        for (auto const& d: data) {
            char message[16];
            sprintf(message, "%0.2f", d.value);
            std::string topic = std::string(DEVICE_NAMESPACE) + std::string(d.name->c_str()) + std::string("/") + std::string(d.type->c_str());

            publish_message(topic, message);
        }
    }
}